# Erxperiment 1
# Sebastian Malkusch, Jörn Lötsch (2021)
# 
# Comparison of substitution options of values outside the limits of quantification
# and their effects on the distribution of the processed variables
# 
# Values from data set #1 that are below the LLOQ are determined and replaced by 
# the LLOQ. The distribution of the substituted variables is then compared to that of
# the original variables to check for changes in the variable’s PDF.
# Values below the LLOQ are substituted with the LLOQ or 0.5LLOQ. Subsequently, the
# PDF of the substituted data is compared with that of the original data by applying
# two-sample Kolmogorov-Smirnov and Wilcoxon's signed rank tests. The precision of
# the imputation is quantified as root mean squared percentage error (RMSPE), using
# the original values as the reference.
# 
# # run time approx. 1 min.


library("tidyverse")
library("ggthemes")
library("cowplot")
library("pguIMP")




extract_missing_values = function(data_df = "tbl.df", imperfect_df = "tbl.df")
{
  if((tibble::is_tibble(data_df)) & (tibble::is_tibble(imperfect_df)))
  {
    # detect missings
    missings <- pguIMP::pgu.missings$new()
    missings$resetImputationParameter(imperfect_df)
    missings_mat <- as.matrix(missings$one_hot_df)
    
    missings_mat[missings_mat < 1] <- NA
    
    # get values of simulated missings
    missings_vector <- (missings_mat * as.matrix(data_df)) %>%
      tibble::as_tibble() %>%
      tidyr::gather(key = "key", value = "value") %>%
      tidyr::drop_na(value) %>%
      dplyr::pull(value)
    return(missings_vector)
  }
  else
  {
    return(NULL)
  }
}

rmspe = function(original_vec = "numeric", measured_vec = "numeric")
{
  p_error =  100*(((original_vec - measured_vec)/original_vec)^2)
  p_error %>%
    mean() %>%
    sqrt() %>%
    return()
}


###############
# import data #
###############
feature_names <- c("S1P", "C16Sphinganin", "C16Cer", "C20Cer", "C24Cer", "C24_1Cer", "C16GluCer", "C16LacCer")
data_df <- readxl::read_xlsx(path = "data/dataset_1.xlsx",
                             sheet = 1) %>%
  dplyr::select(dplyr::all_of(feature_names))

##############
# import loq #
##############
loq_obj <- pguIMP::pgu.limitsOfQuantification$new()
loq_obj$reset(attribute_names = feature_names,
              data_df = readxl::read_xlsx(path = "data/dataset_1.xlsx", sheet = 2))

##########################
# loq outliers detection #
##########################
loq_obj$setNaHandlingAgent <- "keep"
loq_obj$fit(data_df)
print(loq_obj$loqStatistics)

##########################
# Rag and bone variables #
##########################
validation_df <- tibble::tibble(original = dplyr::pull(data_df, "C16Cer"))
value_type_vec <- rep("valid", nrow(validation_df))
value_type_vec[loq_obj$outliers$instance] <- "outlier"
validation_df <- validation_df %>%
  dplyr::mutate(type = value_type_vec)

error_df <- tibble::tibble(method = c("LLOQ", "0.5 LLOQ"))

###########################
# loq outliers prediction #
###########################
loq_obj$setLloqSubstituteAgent <- "keep"
loq_obj$setUloqSubstituteAgent <- "keep"
org_data_df <- loq_obj$predict(data_df)

loq_obj$setLloqSubstituteAgent <- "LLOQ"
loq_obj$setUloqSubstituteAgent <- "ULOQ"
lloq_data_df <- loq_obj$predict(data_df)

loq_obj$setLloqSubstituteAgent <- "0.5 LLOQ"
loq_obj$setUloqSubstituteAgent <- "ULOQ"
half_lloq_data_df <- loq_obj$predict(data_df)

loq_obj$setLloqSubstituteAgent <- "NA"
loq_obj$setUloqSubstituteAgent <- "NA"
imperfect_data_df <- loq_obj$predict(data_df)

validation_df <- validation_df %>%
  dplyr::mutate(LLOQ = lloq_data_df$C16Cer) %>%
  dplyr::mutate("0.5 LLOQ" := half_lloq_data_df$C16Cer)

###################
# detect missings #
###################
org_outliers_vec <- extract_missing_values(data_df = org_data_df, imperfect_df = imperfect_data_df)
lloq_outliers_vec <- extract_missing_values(data_df = lloq_data_df, imperfect_df = imperfect_data_df)
half_lloq_outliers_vec <- extract_missing_values(data_df = half_lloq_data_df, imperfect_df = imperfect_data_df)

rmspe_vec <- c(rmspe(org_outliers_vec, lloq_outliers_vec),rmspe(org_outliers_vec, half_lloq_outliers_vec))
error_df <- error_df %>%
  dplyr::mutate(rmspe = rmspe_vec)

ks_test_obj <- stats::ks.test(x = org_data_df$C16Cer,
                           y = lloq_data_df$C16Cer,
                           alternative = "two.sided",
                           simulate.p.value=TRUE,
                           B=2000,
                           exact = FALSE)

d.Kolmogorow <- c(ks_test_obj$statistic)
p.Kolmogorow <- c(ks_test_obj$p.value)

ks_test_obj <- stats::ks.test(x = org_data_df$C16Cer,
                           y = half_lloq_data_df$C16Cer,
                           alternative = "two.sided",
                           simulate.p.value=TRUE,
                           B=2000,
                           exact = FALSE)

d.Kolmogorow[2] <- ks_test_obj$statistic
p.Kolmogorow[2] <- ks_test_obj$p.value

w_test_obj <- stats::wilcox.test(x = org_data_df$C16Cer,
                           y = lloq_data_df$C16Cer,
                           alternative = "two.sided",
                           exact = FALSE)

w.Wilcoxon <- c(w_test_obj$statistic)
p.Wilcoxon <- c(w_test_obj$p.value)

w_test_obj <- stats::wilcox.test(x = org_data_df$C16Cer,
                           y = half_lloq_data_df$C16Cer,
                           alternative = "two.sided",
                           exact = FALSE)

w.Wilcoxon[2] <- w_test_obj$statistic
p.Wilcoxon[2] <- w_test_obj$p.value

error_df <- error_df %>%
  dplyr::mutate(d.Kolmogorow = d.Kolmogorow) %>%
  dplyr::mutate(p.Kolmogorow = p.Kolmogorow) %>%
  dplyr::mutate(w.Wilcoxon = w.Wilcoxon) %>%
  dplyr::mutate(p.Wilcoxon = p.Wilcoxon)


###############
# save result #
###############
error_df %>%
  writexl::write_xlsx("results/validation_outliers_data.xlsx")

###############
# plot result #
###############
intercept <- loq_obj$loq %>%
  dplyr::filter(attribute == "C16Cer") %>%
  dplyr::pull(LLOQ)
valid_box_plot <- validation_df %>%
  tidyr::gather(key = "key", value = "value", -c("type")) %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x=key, y=value)) +
  ggplot2::geom_boxplot() +
  ggplot2::geom_jitter(mapping = ggplot2::aes(color = type)) +
  ggplot2::xlab("imputation method") +
  ggplot2::ylab("C16Cer values") +
  ggplot2::geom_hline(yintercept = intercept, linetype = "dashed") +
  ggthemes::scale_color_gdocs() +
  ggplot2::theme_bw() +
  ggplot2::theme(axis.text = ggplot2::element_text(size=12),
                 axis.title.x = ggplot2::element_text(size=12,face="bold"),
                 axis.title.y = ggplot2::element_text(size=12,face="bold"),
                 title = ggplot2::element_text(size = 12, face = "bold"),
                 legend.position = "right")
                    
                    

valid_kde_plot <- validation_df %>%
  tidyr::gather(key = "substitute", value = "value", -c("type")) %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x=value, color = substitute)) +
  ggplot2::geom_density(size=1) +
  ggplot2::xlab("C16Cer values") +
  ggplot2::ylab("density") +
  ggplot2::geom_vline(xintercept = intercept, linetype = "dashed") +
  ggthemes::scale_color_gdocs() +
  ggplot2::theme_bw() +
  ggplot2::theme(axis.text = ggplot2::element_text(size=12),
                 axis.title.x = ggplot2::element_text(size=12,face="bold"),
                 axis.title.y = ggplot2::element_text(size=12,face="bold"),
                 title = ggplot2::element_text(size = 12, face = "bold"),
                 legend.position = "right")

p_final <- cowplot::plot_grid(
  valid_box_plot,
  valid_kde_plot,
  ncol = 2,
  labels = c("A", "B"),
  label_size = 15
)

p_final

outfile_name_img <- "results/analyze_loq_outliers.eps"
p_final %>%
  ggplot2::ggsave(filename = outfile_name_img,
                  units = "in",
                  width = 10,
                  height = 5)