# Erxperiment 2
# Sebastian Malkusch, Jörn Lötsch (2021)
# 
# Characterizing the randomness of missing values.
# 
# For the experiment, two incomplete datasets are generated
# from the complete dataset #2 using the “simulateMissings” function
# from the “compositions” package. In each of the data sets, values
# are removed with a probability of p = 0.1 either CAR or NAR. In
# the latter method, low values are removed with higher probability.
# Subsequently, the two data sets are analyzed according to the method
# described in the main text using the “missing_pairs” function of
# the package “finalfit”. The experiment is repeated 100 times and the
# fraction of significantly different distributions per iteration is documented.
# Finally, the documented fractions of significantly different distributions
# in CAR and NAR data sets are compared using the independent 2-group Mann-Whitney U test.
# 
# run time approx. x mins.



library("tidyverse")
library("compositions")
library("ggthemes")
library("cowplot")
library("pguIMP")

#############
# functions #
#############
simulate_missing_values = function(data_df = "tbl.df", mcar=0.0, mnar=0.0, seed_value = 42)
{
  if(tibble::is_tibble(data_df))
  {
    transformator_obj <- pguIMP::pgu.transformator$new()
    normalizer_obj <- pguIMP::pgu.normalizer$new()
    feature_names <- colnames(data_df)
    
    transformator_obj$initialize(data_df)
    for (feature in feature_names){
      transformator_obj$setTrafoType(feature = feature,
                                     type = "tukeyLOP")
      transformator_obj$setMirrorLogic(feature = feature,
                                       logic = FALSE)
    }
    transformator_obj$fit(data_df)
    transformed_df <- transformator_obj$mutateData(data_df)
    
    normalizer_obj$setNormAgent <- "min-max"
    normalizer_obj$detectNormParameter(transformed_df)
    scaled_df <- normalizer_obj$scale_data(transformed_df)
    
    set.seed(seed_value)
    scaled_df %>%
      compositions::simulateMissings(MARprob = mcar, MNARprob = mnar) %>%
      tibble::as_tibble() %>%
      normalizer_obj$rescale_data() %>%
      transformator_obj$reverseMutateData() %>%
      return()
  }
  else
  {
    return(NULL)
  }
}


#############
# load data #
#############
feature_names <- c("MOR", "M3G", "M6G", "COD", "C6G")
data_df <- readxl::read_xlsx(path = "data/dataset_2.xlsx",
                             sheet = 1) %>%
  dplyr::select(dplyr::all_of(feature_names))



##########################
# Rag and bone variables #
##########################
p_missing <- 0.1
max_iter <- 100
mcar_significant_vars <- rep(1.0, max_iter)
mnar_significant_vars <- rep(1.0, max_iter)
seed_vector <- seq(1,max_iter,1)

#####################
# simulate missings #
#####################
for (i in seed_vector){
  seed_int <- as.integer(i)
  mcar_df <- simulate_missing_values(data_df = data_df, mcar=p_missing, seed_value = seed_int)
  mnar_df <- simulate_missing_values(data_df = data_df, mnar=p_missing, seed_value = seed_int)
  
  ######################################
  # characterize missings distribution #
  ######################################
  missings_characterizer_obj <- pguIMP::pgu.missingsCharacterizer$new()
  missings_characterizer_obj$analyze(mcar_df)
  mcar_dist_df <- missings_characterizer_obj$missingsCharacteristics_df %>%
    dplyr::mutate(pValue = stringr::str_replace(pValue, "<", "")) %>%
    dplyr::mutate(pValue = as.numeric(pValue))
  
  missings_characterizer_obj$analyze(mnar_df)
  mnar_dist_df <- missings_characterizer_obj$missingsCharacteristics_df %>%
    dplyr::mutate(pValue = stringr::str_replace(pValue, "<", "")) %>%
    dplyr::mutate(pValue = as.numeric(pValue))

  mcar_significant_vars[i] <- sum(mcar_dist_df$pValue<0.05)
  mnar_significant_vars[i] <- sum(mnar_dist_df$pValue<0.05)
}

dist_df <- tibble::tibble(seed = seed_vector,
                          CAR = mcar_significant_vars/20.0,
                          NAR = mnar_significant_vars/20.0)

sig_dist_plot <- dist_df %>%
  tidyr::gather(key = "missings", value = "significants", -c("seed")) %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x=significants, fill = missings, colour = missings)) +
  ggplot2::geom_density(fill = 0.4, bw = 0.04, size =1.0) +
  ggplot2::ggtitle("Fraction of significant entries from 100 simulations") +
  ggplot2::ylab("density") +
  ggplot2::xlab("fraction of significant entries") +
  ggthemes::scale_color_gdocs() +
  ggplot2::theme_bw() +
  ggplot2::theme(axis.text = ggplot2::element_text(size=12),
                 axis.title.x = ggplot2::element_text(size=12,face="bold"),
                 axis.title.y = ggplot2::element_text(size=12,face="bold"),
                 title = ggplot2::element_text(size = 12, face = "bold"),
                 legend.position = "right")

# independent 2-group Mann-Whitney U Test
wilcox.test(dist_df$CAR,dist_df$NAR)

dist_df %>%
  dplyr::filter(NAR == max(NAR)) %>%
  dplyr::filter(CAR == 0)

# get optimal seed
opt_seed_vect <- dist_df %>%
  dplyr::filter(NAR == max(NAR)) %>%
  dplyr::filter(CAR == 0) %>%
  dplyr::pull(seed)

seed_int <- as.integer(opt_seed_vect[1])

mcar_df <- simulate_missing_values(data_df = data_df, mcar=p_missing, seed_value = seed_int)
mnar_df <- simulate_missing_values(data_df = data_df, mnar=p_missing, seed_value = seed_int)

######################################
# characterize missings distribution #
######################################
missings_characterizer_obj <- pguIMP::pgu.missingsCharacterizer$new()
missings_characterizer_obj$analyze(mcar_df)
mcar_dist_plot <- missings_characterizer_obj$plot_pair_dist(mcar_df)
mcar_dist_df <- missings_characterizer_obj$missingsCharacteristics_df %>%
  dplyr::mutate(pValue = stringr::str_replace(pValue, "<", "")) %>%
  dplyr::mutate(pValue = as.numeric(pValue))

missings_characterizer_obj$analyze(mnar_df)
mnar_dist_plot <- missings_characterizer_obj$plot_pair_dist(mnar_df)
mnar_dist_df <- missings_characterizer_obj$missingsCharacteristics_df %>%
  dplyr::mutate(pValue = stringr::str_replace(pValue, "<", "")) %>%
  dplyr::mutate(pValue = as.numeric(pValue))


##############
# final plot #
##############

top_row <- cowplot::plot_grid(
  GGally::ggmatrix_gtable(mcar_dist_plot + 
                            ggplot2::ggtitle("Distribution matrix with missings simulated CAR")+
                            ggplot2::theme(title = ggplot2::element_text(size = 12, face = "bold"))),
  GGally::ggmatrix_gtable(mnar_dist_plot + 
                            ggplot2::ggtitle("Distribution matrix with missings simulated NAR") +
                            ggplot2::theme(title = ggplot2::element_text(size = 12, face = "bold"))),
  ncol = 2,
  labels = c("A", "B"),
  label_size = 15
)


bottom_row <- cowplot::plot_grid(
  NULL, sig_dist_plot, NULL,
  ncol=3,
  labels = c("", "C", ""),
  rel_widths=c(0.2, 0.6, 0.2),
  label_size = 15
)

p_final <-cowplot::plot_grid(top_row, bottom_row, ncol=1, rel_heights=c(0.7, 0.3))

p_final

outfile_name_img <- "results/characterize_randomness.eps"

p_final %>%
  ggplot2::ggsave(filename = outfile_name_img,
                  units = "in",
                  width = 10,
                  height = 10)