# Erxperiment 3
# Sebastian Malkusch, Jörn Lötsch (2021)
# 
# Validating imputation methods.
# 
# From the complete data set #2, instance values are removed with a
# probability of p = 0.1 using the “simulateMissings” function
# from the “compositions” package as described in the main text.
# The resulting imperfect data sets are transformed according to
# Tukey's ladder of powers with optimized lambda and then min-max normalized.
# Subsequently, missings are substituted by different values:
# The variable mean, the variable median, or values machine-learned
# from the remaining variables using different
# models (CART, kNN, PMM). Finally, the imputed data sets are
# transformed back to their original state and substituted values are
# compared with the original values in the form of the RMSPE.
# The experiment is performed with missings simulated CAR and NAR and
# each procedure is repeated 100 times.
# 
# run time approx. 50 mins.

library("tidyverse")
library("compositions")
library("ggthemes")
library("cowplot")
library("pguIMP")

#############
# functions #
#############
simulate_missing_values = function(data_df = "tbl.df", mcar=0.0, mnar=0.0, seed_value = 42)
{
  if(tibble::is_tibble(data_df))
  {
    transformator_obj <- pguIMP::pgu.transformator$new()
    normalizer_obj <- pguIMP::pgu.normalizer$new()
    feature_names <- colnames(data_df)
    
    transformator_obj$initialize(data_df)
    for (feature in feature_names){
      transformator_obj$setTrafoType(feature = feature,
                                     type = "tukeyLOP")
      transformator_obj$setMirrorLogic(feature = feature,
                                       logic = FALSE)
    }
    transformator_obj$fit(data_df)
    transformed_df <- transformator_obj$mutateData(data_df)
    
    normalizer_obj$setNormAgent <- "min-max"
    normalizer_obj$detectNormParameter(transformed_df)
    scaled_df <- normalizer_obj$scale_data(transformed_df)
    
    set.seed(seed_value)
    scaled_df %>%
      compositions::simulateMissings(MARprob = mcar, MNARprob = mnar) %>%
      tibble::as_tibble() %>%
      normalizer_obj$rescale_data() %>%
      transformator_obj$reverseMutateData() %>%
      return()
  }
  else
  {
    return(NULL)
  }
}

extract_missing_values = function(data_df = "tbl.df", imperfect_df = "tbl.df")
{
  if((tibble::is_tibble(data_df)) & (tibble::is_tibble(imperfect_df)))
  {
    # detect missings
    missings <- pguIMP::pgu.missings$new()
    missings$resetImputationParameter(imperfect_df)
    missings_mat <- as.matrix(missings$one_hot_df)
    
    missings_mat[missings_mat < 1] <- NA
    
    # get values of simulated missings
    missings_vector <- (missings_mat * as.matrix(data_df)) %>%
      tibble::as_tibble() %>%
      tidyr::gather(key = "key", value = "value") %>%
      tidyr::drop_na(value) %>%
      dplyr::pull(value)
    return(missings_vector)
  }
  else
  {
    return(NULL)
  }
}

substitute_missing_values = function(data_df = "tbl.df", method_str = "mean", seed_value = 42)
{
  if(tibble::is_tibble(data_df))
  {
    transformator_obj <- pguIMP::pgu.transformator$new()
    normalizer_obj <- pguIMP::pgu.normalizer$new()
    missings_obj <- pgu.missings$new()
    imputation_obj <- pguIMP::pgu.imputation$new(seed = seed_value,
                                                 iterations = 5,
                                                 imputationAgent = method_str,
                                                 nNeighbors = 4,
                                                 pred_frac = 1.0,
                                                 outflux_thr = 0.5)
    
    feature_names <- colnames(data_df)
    
    transformator_obj$initialize(data_df)
    for (feature in feature_names){
      transformator_obj$setTrafoType(feature = feature,
                                     type = "tukeyLOP")
      transformator_obj$setMirrorLogic(feature = feature,
                                       logic = FALSE)
    }
    transformator_obj$fit(data_df)
    transformed_df <- transformator_obj$mutateData(data_df)
    
    normalizer_obj$setNormAgent <- "min-max"
    normalizer_obj$detectNormParameter(transformed_df)
    scaled_df <- normalizer_obj$scale_data(transformed_df)
    
    missings_obj$resetImputationParameter(scaled_df)
    imputation_obj$gatherImputationSites(missings_df = missings_obj$imputationSites)
    imputation_obj$analyzeImputationSites(data_df = scaled_df)
    imputation_obj$handleImputationSites(data_df = scaled_df) %>%
      normalizer_obj$rescale_data() %>%
      transformator_obj$reverseMutateData() %>%
      return()
  }
  else
  {
    return(NULL)
  }
}

rmspe = function(original_vec = "numeric", measured_vec = "numeric")
{
  p_error =  100*(((original_vec - measured_vec)/original_vec)^2)
  p_error %>%
    mean() %>%
    sqrt() %>%
    return()
}

estimate_substitution_error = function(data_df = "tbl.df",mcar=0.0, mnar=0.0, method_str = "mean", seed_value = 42)
{
  if(tibble::is_tibble(data_df))
  {
    missing_df <- simulate_missing_values(data_df = data_df, mcar=mcar, mnar=mnar, seed_value = seed_value)
    substituted_mcar_df <- substitute_missing_values(data_df = missing_df, method_str = method_str)
    original_missing_values <- extract_missing_values(data_df = data_df, imperfect_df = missing_df)
    substituted_missing_values <- extract_missing_values(data_df = substituted_mcar_df, imperfect_df = missing_df)
    error_value <- rmspe(original_missing_values, substituted_missing_values)
    return(error_value)
  }
  else
  {
    return(NA)
  }
}


#############
# load data #
#############
feature_names <- c("MOR", "M3G", "M6G", "COD", "C6G")
data_df <- readxl::read_xlsx(path = "data/dataset_2.xlsx",
                             sheet = 1) %>%
  dplyr::select(dplyr::all_of(feature_names))



##########################
# Rag and bone variables #
##########################
n_iterations <- 100
p_missing <- 0.1
substitution_error_df <- tibble::tibble(seed_int = c(seq(1,n_iterations),seq(1,n_iterations))) %>%
  dplyr::mutate(mean = c(rep(0.00, n_iterations), rep(p_missing, n_iterations))) %>%
  dplyr::mutate(median = c(rep(0.00, n_iterations), rep(p_missing, n_iterations))) %>%
  dplyr::mutate(cart =c(rep(0.00, n_iterations), rep(p_missing, n_iterations))) %>%
  dplyr::mutate(knn = c(rep(0.00, n_iterations), rep(p_missing, n_iterations))) %>%
  dplyr::mutate(pmm = c(rep(0.00, n_iterations), rep(p_missing, n_iterations))) %>%
  tidyr::gather(key = "method", value = "p_mcar" ,c("mean", "median", "cart", "knn", "pmm")) %>%
  dplyr::mutate(method = factor(method,levels = c("mean", "median", "cart", "knn", "pmm"))) %>%
  dplyr::mutate(p_mnar = p_missing - p_mcar)

####################################
# simulate missings and substitute #
####################################

substitution_error_df <- substitution_error_df %>%
  dplyr::rowwise() %>%
  dplyr::mutate(rmspe = estimate_substitution_error(data_df,  p_mcar, p_mnar, method, seed_int))

#################################
# validate substitution methods #
#################################
mcar_limits <- substitution_error_df %>%
  dplyr::filter(p_mnar < p_missing) %>%
  dplyr::pull(rmspe) %>%
  quantile(c(0.1,0.85), na.rm = TRUE)
mcar_limits[1] <- 0

mnar_limits <- substitution_error_df %>%
  dplyr::filter(p_mcar < p_missing) %>%
  dplyr::pull(rmspe) %>%
  quantile(c(0.1,0.96), na.rm=TRUE)
mnar_limits[1] <- 0

mcar_substitute_plot <- substitution_error_df %>%
  dplyr::filter(p_mnar < p_missing) %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x=method, y=rmspe)) +
  ggplot2::geom_boxplot(outlier.shape = NA) +
  ggplot2::coord_cartesian(ylim = mcar_limits) +
  ggplot2::ggtitle("Missings simulated CAR") +
  ggplot2::ylab("RMSPE [%]") +
  ggplot2::xlab("substitution method") +
  ggplot2::theme_bw() +
  ggplot2::theme(axis.text = ggplot2::element_text(size=12),
                 axis.title.x = ggplot2::element_text(size=12,face="bold"),
                 axis.title.y = ggplot2::element_text(size=12,face="bold"),
                 title = ggplot2::element_text(size = 12, face = "bold"),
                 legend.position = "none")

mnar_substitute_plot <- substitution_error_df %>%
  dplyr::filter(p_mcar < p_missing) %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x=method, y=rmspe)) +
  ggplot2::geom_boxplot(outlier.shape = NA) +
  ggplot2::geom_jitter() +
  ggplot2::coord_cartesian(ylim = mnar_limits) +
  ggplot2::ggtitle("Missings simulated NAR") +
  ggplot2::ylab("RMSPE [%]") +
  ggplot2::xlab("substitution method") +
  ggplot2::theme_bw() +
  ggplot2::theme(axis.text = ggplot2::element_text(size=12),
                 axis.title.x = ggplot2::element_text(size=12,face="bold"),
                 axis.title.y = ggplot2::element_text(size=12,face="bold"),
                 title = ggplot2::element_text(size = 12, face = "bold"),
                 legend.position = "none")

##############
# final plot #
##############

substitution_error_df %>%
  writexl::write_xlsx("results/validate_imputation_methods.xlsx")

p_final <- cowplot::plot_grid(
  mcar_substitute_plot,
  mnar_substitute_plot,
  ncol=2,
  labels = c("A", "B"),
  label_size = 15
)

p_final

outfile_name_img <- "results/validate_imputation_methods.eps"

p_final %>%
  ggplot2::ggsave(filename = outfile_name_img,
                  units = "in",
                  width = 10,
                  height = 3)