# Erxperiment 4 and 5
# Sebastian Malkusch, Jörn Lötsch (2021)
# 
# Assessment of the impact of data imputation on high-dimensional data structures.
# 
# To compare imputation methods in terms of their effect on projecting high-dimensional
# data onto a low-dimensional plane, two sample data sets are created from data set #3
# by substituting outliers. First, data are Ln(x) transformed and min-max normalized.
# Second, imputation sites are introduced by replacing outliers identified with Grubb's
# tests (alpha=0.05) with missing values. Third, missing values are imputed either by
# feature medians or using a kNN model trained on the remaining instances.
# The transformed and normalized original data served as control.
# Dimensionality reduction is achieved by means of principal component analysis (PCA)
# on centered data using the "prcomp" method of the basic R package "stats".
# The first two principal components are kept. To quantify the loss of information due
# to the dimensionality reduction process,the projected data are transformed back from
# the two-dimensional space to their original coordinate space and the reconstruction
# error is quantified using the mean squared error (MSE) as metric.
# 
# Estimation of consequences of outlier imputation for data set subgroup structures
# 
# Consequences of outliers and their imputation are further evaluated in the context
# of clustering as a common task in biomedical data analysis to identify subgroups.
# For this purpose, a group structure is sought in the three versions of the dimensionality
# reduced data set #3 using centroid- and density-based methods of unsupervised learning.
# The results are compared with the known three-group structure described in the main text.
# Centroid-based clustering is realized by training a k-means clustering model (k-means)
# on the XPCA data using the “kmeans” function of the “stats” R base package.
# Graphical validation of the procedure is done with a scatter plot of the XPCA data
# color coded by the true underlying group structure, superimposing the proposed groups
# as convex hulls. Alternatively, density-based clustering is performed by training
# the OPTICS algorithm implemented in the “dbscan” package on the XPCA data
# The hierarchical clustering structure is visualized as dendrogram and reachability
# distance plot. Subsequently, the DBSCAN cluster structure is extracted from the OPTICS
# result by defining a reachability-distance threshold that would result in a clustering
# solution matching the true number of clusters using the “extractDBSCAN” function o
# the “dbscan” package. Graphical data representation was realized using the “ggplot2” package
# which is part of the “tidyverse” package.
# 
# run time approx. 1 min.

library("tidyverse")
library("ggthemes")
library("ggdendro")
library("gridExtra")
library("dbscan")
library("pguIMP")

########################
# clustering functions #
########################
plot_clusteing_result <- function(cluster_df, group_name, variance_pc1, variance_pc2){
  title_str = sprintf("Clusters by %s", group_name)

  convex_hull <- cluster_df %>%
    dplyr::filter(!!as.symbol(group_name) != 0) %>%
    dplyr::group_by(.dots = group_name) %>%
    dplyr::slice(grDevices::chull(PC1, PC2))


  convex_hull %>%
    ggplot2::ggplot(mapping = ggplot2::aes(x=PC1, y=PC2)) +
    ggplot2::geom_polygon( mapping = ggplot2::aes(group = !!as.symbol(group_name)), fill = NA, colour="black", linetype = "solid", size = 1)+
    ggplot2::geom_point(data = cluster_df, mapping = ggplot2::aes(fill = cluster, stroke = measurement_type), color = "black", shape = 21, size = 4) +
    ggplot2::scale_discrete_manual(aesthetics = "stroke", values = c(1.5,0)) +
    ggplot2::geom_text(data = cluster_df, mapping = ggplot2::aes(label=id, color = measurement_type),hjust=1, vjust=-2) +
    #ggplot2::scale_color_manual(values =ggthemes::gdocs_pal()(10)[c(1,4,5)], name = "Label") +
    ggplot2::scale_color_manual(values = c("black", "dark gray")) +
    ggplot2::scale_fill_manual(values =ggthemes::gdocs_pal()(10)[c(1,4,5)], name = "Label") +
    ggplot2::ggtitle(title_str) +
    ggplot2::xlab(sprintf("PC1 (%.2f%%)", variance_pc1*100)) +
    ggplot2::ylab(sprintf("PC2 (%.2f%%)", variance_pc2*100)) +
    ggplot2::xlim(c(-1.0, 1.0)) +
    ggplot2::ylim(c(-1.0, 1.0)) +
    ggplot2::theme_bw() +
    ggplot2::theme(axis.text = ggplot2::element_text(size=12),
                   axis.title.x = ggplot2::element_text(size=12,face="bold"),
                   axis.title.y = ggplot2::element_text(size=12,face="bold"),
                   title = ggplot2::element_text(size = 12, face = "bold"),
                   legend.position = "none") %>%
    return()

}


plot_optics_dendrogram <- function(res_optics, cluster_df, thr){
  cluster_df <- cluster_df %>%
    dplyr::select(c("id", "order", "cluster", "reach_dist")) %>%
    dplyr::mutate(cluster = cluster_df$cluster[cluster_df$order]) %>%
    dplyr::mutate(measurement_type = cluster_df$measurement_type[cluster_df$order])

  y_max <- max(sort(cluster_df$reach_dist, decreasing = TRUE)[-1])

  res_optics %>%
    dbscan::as.reachability() %>%
    stats::as.dendrogram() %>%
    ggdendro::dendro_data() %>%
    ggdendro::segment() %>%
    ggplot2::ggplot()+
    ggplot2::geom_segment(aes(x=x, y=y, xend=xend, yend=yend)) +
    ggplot2::geom_hline(yintercept = thr, linetype = "dashed") +
    ggplot2::geom_label(data = cluster_df, mapping = ggplot2::aes(x = id, y = -0.05, label = order, fill = cluster, color = measurement_type), size = 2, fontface = "bold") +
    ggplot2::scale_color_manual(values = c("black", "dark gray")) +
    ggplot2::scale_fill_manual(values =ggthemes::gdocs_pal()(10)[c(1,4,5)], name = "Label") +
    ggplot2::xlab("instance ID") +
    ggplot2::ylab("reachability distance") +
    ggplot2::xlim(c(0, 19)) +
    ggplot2::ylim(c(-0.1, y_max)) +
    ggplot2::ggtitle("OPTICS dendrogram") +
    ggplot2::theme_bw() +
    ggplot2::theme(axis.text.x = ggplot2::element_blank(),
                   axis.text.y = ggplot2::element_text(size=12),
                   axis.title.x = ggplot2::element_text(size=12,face="bold"),
                   axis.title.y = ggplot2::element_text(size=12,face="bold"),
                   axis.ticks.x = ggplot2::element_blank(),
                   panel.grid.major.x = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank(),
                   panel.border = ggplot2::element_blank(),
                   title = ggplot2::element_text(size = 12, face = "bold"),
                   legend.position = "none") %>%
    return()
}


plot_optics_reachability_vs_order <- function(cluster_df, thr){
  y_max <-   max(sort(cluster_df$reach_dist, decreasing = TRUE)[-1])

  optics_df <- tibble::tibble(order = cluster_df$id,
                              id = cluster_df$order,
                              reach_dist = cluster_df$reach_dist[cluster_df$order],
                              cluster = cluster_df$cluster[cluster_df$order],
                              measurement_type = cluster_df$measurement_type[cluster_df$order])

  optics_df %>%
    ggplot2::ggplot(mapping = ggplot2::aes(x=order, y=reach_dist, fill = cluster)) +
    ggplot2::geom_col() +
    ggplot2::scale_fill_manual(values =ggthemes::gdocs_pal()(10)[c(1,4,5)], name = "Label") +
    ggplot2::geom_text(aes(label=id, color = measurement_type), vjust=-0.25) +
    ggplot2::scale_color_manual(values = c("black", "dark gray")) +
    ggplot2::geom_hline(yintercept = thr, linetype = "dashed") +
    ggplot2::xlab("instance order") +
    ggplot2::ylab("reachability distance") +
    ggplot2::ylim(c(0, y_max + 0.1)) +
    ggplot2::ggtitle("OPTICS reachability") +
    ggplot2::theme_bw() +
    ggplot2::theme(axis.text = ggplot2::element_text(size=12),
                   axis.title.x = ggplot2::element_text(size=12,face="bold"),
                   axis.title.y = ggplot2::element_text(size=12,face="bold"),
                   title = ggplot2::element_text(size = 12, face = "bold"),
                   legend.position = "none") %>%
    return()
}
#################
# pca functions #
#################
pca.recon <- function(pca_res, k){
  mu <- matrix(rep(pca_res$center, nrow(pca_res$x)), nrow = nrow(pca_res$x), byrow = T)
  recon <- pca_res$x[,1:k] %*% t(pca_res$rotation[,1:k]) + mu
  return(recon)
}



#################
# mse functions #
#################
square <- function(data_df){
  return(data_df^2)
}

calc.mse <- function(org_df, recon_df){
  delta_df <- org_df - recon_df
  delta_df %>%
    as.matrix() %>%
    square() %>%
    mean(na.rm = TRUE) %>%
    return()
}



##############
# file names #
##############
folder_str <- sprintf("data")
folder_image_str <- sprintf("results")
file_str <- sprintf("dataset_3.csv")


########################
# init process objects #
########################
file_name_obj <- pguIMP::pgu.file$new()
importer_obj <- pguIMP::pgu.importer$new()
loq_obj <- pguIMP::pgu.limitsOfQuantification$new()
filter_obj <- pguIMP::pgu.filter$new()
transformator_obj <- pguIMP::pgu.transformator$new()
transform_model_obj <- pguIMP::pgu.model$new()
normalizer_obj <- pguIMP::pgu.normalizer$new()
missings_obj <- pguIMP::pgu.missings$new()
outliers_obj <- pguIMP::pgu.outliers$new()
imputer_obj <- pguIMP::pgu.imputation$new()
validator_obj <- pguIMP::pgu.validator$new()
transform_model_obj <- pguIMP::pgu.model$new()

#####################
# init data objects #
#####################
raw_data <- pguIMP::pgu.data$new()
meta_data <- pguIMP::pgu.data$new()
filtered_data <- pguIMP::pgu.data$new()
transformed_data <- pguIMP::pgu.data$new()
normalized_data <- pguIMP::pgu.data$new()
knn_imputed_data <- pguIMP::pgu.data$new()
median_imputed_data <- pguIMP::pgu.data$new()
mean_imputed_data <- pguIMP::pgu.data$new()
weka_imputed_data <- pguIMP::pgu.data$new()


##########################
# Rag and bone variables #
##########################
set.seed(42)
org_model_quality_df <- tibble::tibble()
knn_imp_model_quality_df <- tibble::tibble()
median_imp_model_quality_df <- tibble::tibble()
mse_reconstruction_df <- tibble::tibble(data = c("original", "knn_imputed", "median_imputed", "mean_imputed", "weka_imputed"))
org_cluster_df <- tibble::tibble()
knn_cluster_df <- tibble::tibble()
median_cluster_df <- tibble::tibble()
org_dbscan_thr <- 0.0
knn_dbscan_thr <- 0.0
median_dbscan_thr <- 0-0
###############
# import data #
###############
file_name_obj$reset(uploadFileName = paste(folder_str, file_str, sep = "/"),
                    fileName = file_str,
                    sheetIndex = 1,
                    separator = ",",
                    skipRows = 0,
                    columnNames = TRUE)
file_name_obj$fit()

raw_data$setRawData <- importDataSet(file_name_obj)

org_cluster_df <- tibble::tibble(id = as.integer(dplyr::pull(raw_data$rawData, "Sample Name")),
                                 cluster = factor(substr(raw_data$rawData$Diagnosis,1,1)))
knn_cluster_df <- tibble::tibble(id = as.integer(dplyr::pull(raw_data$rawData, "Sample Name")),
                                 cluster = factor(substr(raw_data$rawData$Diagnosis,1,1)))
median_cluster_df <- tibble::tibble(id = as.integer(dplyr::pull(raw_data$rawData, "Sample Name")),
                                    cluster = factor(substr(raw_data$rawData$Diagnosis,1,1)))
mean_cluster_df <- tibble::tibble(id = as.integer(dplyr::pull(raw_data$rawData, "Sample Name")),
                                  cluster = factor(substr(raw_data$rawData$Diagnosis,1,1)))
weka_cluster_df <- tibble::tibble(id = as.integer(dplyr::pull(raw_data$rawData, "Sample Name")),
                                  cluster = factor(substr(raw_data$rawData$Diagnosis,1,1)))

###############
# filter data #
###############
filter_obj$reset(data = raw_data$rawData)
filter_feature_list <- c("lipid_mediator_2",
                         "lipid_mediator_10",
                         "lipid_mediator_140",
                         "lipid_mediator_170",
                         "lipid_mediator_171",
                         "lipid_mediator_175")
filter_idx_list <- which(colnames(raw_data$rawData) %in% filter_feature_list)
filter_obj$setColIdx <- filter_idx_list
filtered_data$setRawData <-  filter_obj$predict(data_df = raw_data$rawData)
##############################
# define transformation type #
##############################
transformator_obj$resetTrafoParameter(filtered_data$numerical_data())
for (feature in filtered_data$numericalAttributeNames){
  transformator_obj$setTrafoType(feature = feature,
                                 type = "logNorm")
  transformator_obj$setMirrorLogic(feature = feature,
                                   logic = FALSE)
}#for
print(transformator_obj)
#############################
# fit transformation object #
#############################
transformator_obj$fit(filtered_data$numerical_data())

##########################
# predict transfromation #
##########################
name  <- as.name("Sample Name")
transformed_data$setRawData <- filtered_data$numerical_data() %>%
  transformator_obj$mutateData() %>%
  tibble::add_column(!! name := filtered_data$rawData %>%
                       dplyr::select(!!name) %>%
                       unlist() %>%
                       as.character()) %>%
  dplyr::select(c(!!name, filtered_data$numericalAttributeNames))

########################
# model transformation #
########################
transform_model_obj$resetModel(transformed_data$numerical_data())
transform_model_obj$modelParameter %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x = p.kolmogorow)) +
  ggplot2::geom_density()

org_model_quality_df <- transform_model_obj$modelParameter %>%
  dplyr::select(c("features", "d.kolmogorow","p.kolmogorow", "a.anderson", "p.anderson", "w.shapiro", "p.shapiro"))

#########################
# fit normalizer object #
#########################
normalizer_obj$setNormAgent <- "min-max"
normalizer_obj$detectNormParameter(transformed_data$numerical_data())

#########################
# predict normalization #
#########################
normalized_data$setRawData <- transformed_data$numerical_data() %>%
  normalizer_obj$scale_data() %>%
  tibble::add_column(!! name := transformed_data$rawData %>%
                       dplyr::select(!!name) %>%
                       unlist() %>%
                       as.character()) %>%
  dplyr::select(c(!!name, transformed_data$numericalAttributeNames))

###########################
# detect imputation sites #
###########################
missings_obj$resetImputationParameter(normalized_data$numerical_data())
outliers_obj$setOutliersAgent <- "Grubbs"
outliers_obj$setAlpha <- 0.05
outliers_obj$detectOutliers(normalized_data$numerical_data())


outliers_obj$outliersStatistics %>%
  dplyr::arrange(dplyr::desc(outlierFraction))


measurement_type <-rep("regular", nrow(knn_cluster_df))
measurement_type[as.integer(outliers_obj$outliers$measurement)] <- "outlier"

org_cluster_df <- org_cluster_df %>%
  dplyr::mutate(measurement_type = measurement_type)
knn_cluster_df <- knn_cluster_df %>%
  dplyr::mutate(measurement_type = measurement_type)
median_cluster_df <- median_cluster_df %>%
  dplyr::mutate(measurement_type = measurement_type)
mean_cluster_df <- mean_cluster_df %>%
  dplyr::mutate(measurement_type = measurement_type)
weka_cluster_df <- weka_cluster_df %>%
  dplyr::mutate(measurement_type = measurement_type)
#########################
# fit imputation object #
#########################
imputer_obj$setImputationAgent <- "knn"
imputer_obj$setNNeighbors <- 2
imputer_obj$setSeed <- 42
imputer_obj$setIterations <- 10
imputer_obj$gatherImputationSites(missings_df = missings_obj$imputationSites,
                                  outliers_df = outliers_obj$outliers)

imputer_obj$analyzeImputationSites(data_df = normalized_data$numerical_data())

imputer_obj$detectPredictors(data_df = normalized_data$numerical_data())

############################
# predict imputation sites #
############################
knn_imputed_data$setRawData <- normalized_data$numerical_data() %>%
  imputer_obj$handleImputationSites() %>%
  tibble::add_column(!! name := normalized_data$rawData %>%
                       dplyr::select(!!name) %>%
                       unlist() %>%
                       as.character()) %>%
  dplyr::select(c(!!name, normalized_data$numericalAttributeNames))

imputer_obj$setImputationAgent <- "median"
median_imputed_data$setRawData <- normalized_data$numerical_data() %>%
  imputer_obj$handleImputationSites() %>%
  tibble::add_column(!! name := normalized_data$rawData %>%
                       dplyr::select(!!name) %>%
                       unlist() %>%
                       as.character()) %>%
  dplyr::select(c(!!name, normalized_data$numericalAttributeNames))

imputer_obj$setImputationAgent <- "mean"
mean_imputed_data$setRawData <- normalized_data$numerical_data() %>%
  imputer_obj$handleImputationSites() %>%
  tibble::add_column(!! name := normalized_data$rawData %>%
                       dplyr::select(!!name) %>%
                       unlist() %>%
                       as.character()) %>%
  dplyr::select(c(!!name, normalized_data$numericalAttributeNames))

##########################
# model knn imputed data #
##########################
knn_imputed_data$numerical_data() %>%
  normalizer_obj$rescale_data() %>%
  transform_model_obj$resetModel()

transform_model_obj$modelParameter %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x = p.kolmogorow)) +
  ggplot2::geom_density()

knn_imp_model_quality_df <- transform_model_obj$modelParameter %>%
  dplyr::select(c("features", "d.kolmogorow","p.kolmogorow", "a.anderson", "p.anderson", "w.shapiro", "p.shapiro"))

#############################
# model median imputed data #
#############################
median_imputed_data$numerical_data() %>%
  normalizer_obj$rescale_data() %>%
  transform_model_obj$resetModel()

transform_model_obj$modelParameter %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x = p.kolmogorow)) +
  ggplot2::geom_density()

median_imp_model_quality_df <- transform_model_obj$modelParameter %>%
  dplyr::select(c("features", "d.kolmogorow","p.kolmogorow", "a.anderson", "p.anderson", "w.shapiro", "p.shapiro"))

#############################
# model mean imputed data #
#############################
mean_imputed_data$numerical_data() %>%
  normalizer_obj$rescale_data() %>%
  transform_model_obj$resetModel()

transform_model_obj$modelParameter %>%
  ggplot2::ggplot(mapping = ggplot2::aes(x = p.kolmogorow)) +
  ggplot2::geom_density()

mean_imp_model_quality_df <- transform_model_obj$modelParameter %>%
  dplyr::select(c("features", "d.kolmogorow","p.kolmogorow", "a.anderson", "p.anderson", "w.shapiro", "p.shapiro"))

#######
# pca #
#######
org_pca_obj <- normalized_data$numerical_data() %>%
  prcomp(center = TRUE, scale. = FALSE)
org_pca_summary <- summary(org_pca_obj)
org_pca_variance <- rep(0.0,2)
org_pca_variance[1] <- org_pca_summary$importance["Proportion of Variance","PC1"]
org_pca_variance[2] <- org_pca_summary$importance["Proportion of Variance","PC2"]
org_cluster_df <- org_cluster_df %>%
  dplyr::mutate(PC1 = org_pca_obj$x[,1],
                PC2 = org_pca_obj$x[,2])

knn_imp_pca_obj <- knn_imputed_data$numerical_data() %>%
  prcomp(center = TRUE, scale. = FALSE)
knn_imp_pca_summary <- summary(knn_imp_pca_obj)
knn_imp_pca_variance <- rep(0.0,2)
knn_imp_pca_variance[1] <- knn_imp_pca_summary$importance["Proportion of Variance","PC1"]
knn_imp_pca_variance[2] <- knn_imp_pca_summary$importance["Proportion of Variance","PC2"]
knn_cluster_df <- knn_cluster_df %>%
  dplyr::mutate(PC1 = knn_imp_pca_obj$x[,1],
                PC2 = knn_imp_pca_obj$x[,2])

median_imp_pca_obj <- median_imputed_data$numerical_data() %>%
  prcomp(center = TRUE, scale. = FALSE)
median_imp_pca_summary <- summary(median_imp_pca_obj)
median_imp_pca_variance <- rep(0.0,2)
median_imp_pca_variance[1] <- median_imp_pca_summary$importance["Proportion of Variance","PC1"]
median_imp_pca_variance[2] <- median_imp_pca_summary$importance["Proportion of Variance","PC2"]
median_cluster_df <- median_cluster_df %>%
  dplyr::mutate(PC1 = median_imp_pca_obj$x[,1],
                PC2 = median_imp_pca_obj$x[,2])

mean_imp_pca_obj <- mean_imputed_data$numerical_data() %>%
  prcomp(center = TRUE, scale. = FALSE)
mean_imp_pca_summary <- summary(mean_imp_pca_obj)
mean_imp_pca_variance <- rep(0.0,2)
mean_imp_pca_variance[1] <- mean_imp_pca_summary$importance["Proportion of Variance","PC1"]
mean_imp_pca_variance[2] <- mean_imp_pca_summary$importance["Proportion of Variance","PC2"]
mean_cluster_df <- mean_cluster_df %>%
  dplyr::mutate(PC1 = mean_imp_pca_obj$x[,1],
                PC2 = mean_imp_pca_obj$x[,2])

pca_variance_df <- tibble::tibble(PC = seq(2),
                                  original = org_pca_variance,
                                  knn_imputed = knn_imp_pca_variance,
                                  median_imputed = median_imp_pca_variance,
                                  mean_imputed = mean_imp_pca_variance)

############################
# reconstruction error pca #
############################
k <- 2
org_pca_mse <- pca.recon(org_pca_obj, k) %>%
  calc.mse(normalized_data$numerical_data())

knn_imp_pca_mse <- pca.recon(knn_imp_pca_obj, k) %>%
  calc.mse(knn_imputed_data$numerical_data())

median_imp_pca_mse <- pca.recon(median_imp_pca_obj, k) %>%
  calc.mse(median_imputed_data$numerical_data())

mean_imp_pca_mse <- pca.recon(mean_imp_pca_obj, k) %>%
  calc.mse(mean_imputed_data$numerical_data())

mse_reconstruction_df <- mse_reconstruction_df %>%
  dplyr::mutate(pca.mse =c(org_pca_mse, knn_imp_pca_mse, median_imp_pca_mse, mean_imp_pca_mse))

##########
# OPTICS #
##########
set.seed(42)
res_optics_org <- org_pca_obj$x %>%
  tibble::as_tibble() %>%
  dplyr::select(PC1, PC2) %>%
  dbscan::optics(eps = 2, minPts = 3)

org_cluster_df <- org_cluster_df %>%
  dplyr::mutate(order = res_optics_org$order,
                reach_dist = res_optics_org$reachdist,
                core_dist = res_optics_org$coredist)

set.seed(42)
res_optics_knn <- knn_imp_pca_obj$x %>%
  tibble::as_tibble() %>%
  dplyr::select(PC1, PC2) %>%
  dbscan::optics(eps = 2, minPts = 3)

knn_cluster_df <- knn_cluster_df %>%
  dplyr::mutate(order = res_optics_knn$order,
                reach_dist = res_optics_knn$reachdist,
                core_dist = res_optics_knn$coredist)


set.seed(42)
res_optics_median <- median_imp_pca_obj$x %>%
  tibble::as_tibble() %>%
  dplyr::select(PC1, PC2) %>%
  dbscan::optics(eps = 2, minPts = 3)

median_cluster_df <- median_cluster_df %>%
  dplyr::mutate(order = res_optics_median$order,
                reach_dist = res_optics_median$reachdist,
                core_dist = res_optics_median$coredist)


set.seed(42)
res_optics_mean <- mean_imp_pca_obj$x %>%
  tibble::as_tibble() %>%
  dplyr::select(PC1, PC2) %>%
  dbscan::optics(eps = 2, minPts = 3)

mean_cluster_df <- mean_cluster_df %>%
  dplyr::mutate(order = res_optics_mean$order,
                reach_dist = res_optics_mean$reachdist,
                core_dist = res_optics_mean$coredist)
##########
# DBSCAN #
##########
org_dbscan_thr <- 0.3
median_dbscan_thr <- 0.24
knn_dbscan_thr <- 0.4
mean_dbscan_thr <- 0.235


res_dbscan_org <- dbscan::extractDBSCAN(res_optics_org, eps_cl = org_dbscan_thr)
org_cluster_df <- org_cluster_df %>%
  dplyr::mutate(DBSCAN = factor(res_dbscan_org$cluster))

res_dbscan_knn <- dbscan::extractDBSCAN(res_optics_knn, eps_cl = knn_dbscan_thr)
knn_cluster_df <- knn_cluster_df %>%
  dplyr::mutate(DBSCAN = factor(res_dbscan_knn$cluster))

res_dbscan_median <- dbscan::extractDBSCAN(res_optics_median, eps_cl = median_dbscan_thr)
median_cluster_df <- median_cluster_df %>%
  dplyr::mutate(DBSCAN = factor(res_dbscan_median$cluster))

res_dbscan_mean <- dbscan::extractDBSCAN(res_optics_mean, eps_cl = mean_dbscan_thr)
mean_cluster_df <- mean_cluster_df %>%
  dplyr::mutate(DBSCAN = factor(res_dbscan_mean$cluster))

##########
# kMeans #
##########
set.seed(42)
res_kmeans_org <- org_pca_obj$x %>%
  tibble::as_tibble() %>%
  dplyr::select(PC1, PC2) %>%
  stats::kmeans(centers = 3)

org_cluster_df <- org_cluster_df %>%
  dplyr::mutate(kMeans = factor(res_kmeans_org$cluster))

set.seed(42)
res_kmeans_knn <- knn_imp_pca_obj$x %>%
  tibble::as_tibble() %>%
  dplyr::select(PC1, PC2) %>%
  stats::kmeans(centers = 3)

knn_cluster_df <- knn_cluster_df %>%
  dplyr::mutate(kMeans = factor(res_kmeans_knn$cluster))

set.seed(42)
res_kmeans_median <- median_imp_pca_obj$x %>%
  tibble::as_tibble() %>%
  dplyr::select(PC1, PC2) %>%
  stats::kmeans(centers = 3)

median_cluster_df <- median_cluster_df %>%
  dplyr::mutate(kMeans = factor(res_kmeans_median$cluster))

set.seed(42)
res_kmeans_mean <- mean_imp_pca_obj$x %>%
  tibble::as_tibble() %>%
  dplyr::select(PC1, PC2) %>%
  stats::kmeans(centers = 3)

mean_cluster_df <- mean_cluster_df %>%
  dplyr::mutate(kMeans = factor(res_kmeans_mean$cluster))

####################
# individual plots #
####################
p_dendrogram_org <- plot_optics_dendrogram(res_optics_org, org_cluster_df, thr = org_dbscan_thr)
p_reachability_org <- plot_optics_reachability_vs_order(org_cluster_df, thr = org_dbscan_thr)
p_dendrogram_knn <- plot_optics_dendrogram(res_optics_knn, knn_cluster_df, thr = knn_dbscan_thr)
p_reachability_knn <- plot_optics_reachability_vs_order(knn_cluster_df, thr = knn_dbscan_thr)

p_dendrogram_median <- plot_optics_dendrogram(res_optics_median, median_cluster_df, thr = median_dbscan_thr)
p_reachability_median <- plot_optics_reachability_vs_order(median_cluster_df, thr = median_dbscan_thr)

p_dendrogram_mean <- plot_optics_dendrogram(res_optics_mean, mean_cluster_df, thr = mean_dbscan_thr)
p_reachability_mean <- plot_optics_reachability_vs_order(mean_cluster_df, thr = mean_dbscan_thr)


p_labeling_org <- org_cluster_df %>%
  dplyr::mutate(labeling = cluster) %>%
  plot_clusteing_result("labeling",
                        org_pca_summary$importance["Proportion of Variance","PC1"],
                        org_pca_summary$importance["Proportion of Variance","PC2"])

org_cluster_df %>%
  dplyr::mutate(labeling = cluster)
p_dbscan_org <- plot_clusteing_result(org_cluster_df,
                                      "DBSCAN",
                                      org_pca_summary$importance["Proportion of Variance","PC1"],
                                      org_pca_summary$importance["Proportion of Variance","PC2"])

p_kmeans_org <- plot_clusteing_result(org_cluster_df,
                                      "kMeans",
                                      org_pca_summary$importance["Proportion of Variance","PC1"],
                                      org_pca_summary$importance["Proportion of Variance","PC2"])

p_labeling_knn <- knn_cluster_df %>%
  dplyr::mutate(labeling = cluster) %>%
  plot_clusteing_result("labeling",
                        knn_imp_pca_summary$importance["Proportion of Variance","PC1"],
                        knn_imp_pca_summary$importance["Proportion of Variance","PC2"])

p_dbscan_knn <- plot_clusteing_result(knn_cluster_df,
                                      "DBSCAN",
                                      knn_imp_pca_summary$importance["Proportion of Variance","PC1"],
                                      knn_imp_pca_summary$importance["Proportion of Variance","PC2"])

p_kmeans_knn <- plot_clusteing_result(knn_cluster_df,
                                      "kMeans",
                                      knn_imp_pca_summary$importance["Proportion of Variance","PC1"],
                                      knn_imp_pca_summary$importance["Proportion of Variance","PC2"])


p_labeling_median <- median_cluster_df %>%
  dplyr::mutate(labeling = cluster) %>%
  plot_clusteing_result("labeling",
                        median_imp_pca_summary$importance["Proportion of Variance","PC1"],
                        median_imp_pca_summary$importance["Proportion of Variance","PC2"])

p_dbscan_median <- plot_clusteing_result(median_cluster_df,
                                         "DBSCAN",
                                         median_imp_pca_summary$importance["Proportion of Variance","PC1"],
                                         median_imp_pca_summary$importance["Proportion of Variance","PC2"])

p_kmeans_median <- plot_clusteing_result(median_cluster_df,
                                         "kMeans",
                                         median_imp_pca_summary$importance["Proportion of Variance","PC1"],
                                         median_imp_pca_summary$importance["Proportion of Variance","PC2"])

p_labeling_mean <- mean_cluster_df %>%
  dplyr::mutate(labeling = cluster) %>%
  plot_clusteing_result("labeling",
                        mean_imp_pca_summary$importance["Proportion of Variance","PC1"],
                        mean_imp_pca_summary$importance["Proportion of Variance","PC2"])

p_dbscan_mean <- plot_clusteing_result(mean_cluster_df,
                                       "DBSCAN",
                                       mean_imp_pca_summary$importance["Proportion of Variance","PC1"],
                                       mean_imp_pca_summary$importance["Proportion of Variance","PC2"])

p_kmeans_mean <- plot_clusteing_result(mean_cluster_df,
                                       "kMeans",
                                       mean_imp_pca_summary$importance["Proportion of Variance","PC1"],
                                       mean_imp_pca_summary$importance["Proportion of Variance","PC2"])


#################
# compound plot #
#################
p1 <- gridExtra::arrangeGrob(p_labeling_org,
                             top = grid::textGrob("A", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))

p2 <- gridExtra::arrangeGrob(p_kmeans_org,
                             top = grid::textGrob("B", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))

p3 <- gridExtra::arrangeGrob(p_dendrogram_org ,
                             top = grid::textGrob("C", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))

p4 <- gridExtra::arrangeGrob(p_reachability_org ,
                             top = grid::textGrob("D", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))


p5 <- gridExtra::arrangeGrob(p_dbscan_org,
                             top = grid::textGrob("E", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))

p6 <- gridExtra::arrangeGrob(p_labeling_median,
                             top = grid::textGrob("F", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))

p7 <- gridExtra::arrangeGrob(p_kmeans_median,
                             top = grid::textGrob("G", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))

p8 <- gridExtra::arrangeGrob(p_dendrogram_median ,
                             top = grid::textGrob("H", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))

p9 <- gridExtra::arrangeGrob(p_reachability_median ,
                             top = grid::textGrob("I", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                  just=c("left","top"),
                                                  gp=grid::gpar(col="black", fontsize=20)))

p10 <- gridExtra::arrangeGrob(p_dbscan_median,
                              top = grid::textGrob("J", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                   just=c("left","top"),
                                                   gp=grid::gpar(col="black", fontsize=20)))

p11 <- gridExtra::arrangeGrob(p_labeling_knn,
                              top = grid::textGrob("K", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                   just=c("left","top"),
                                                   gp=grid::gpar(col="black", fontsize=20)))

p12 <- gridExtra::arrangeGrob(p_kmeans_knn,
                              top = grid::textGrob("L", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                   just=c("left","top"),
                                                   gp=grid::gpar(col="black", fontsize=20)))

p13 <- gridExtra::arrangeGrob(p_dendrogram_knn ,
                              top = grid::textGrob("M", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                   just=c("left","top"),
                                                   gp=grid::gpar(col="black", fontsize=20)))

p14 <- gridExtra::arrangeGrob(p_reachability_knn ,
                              top = grid::textGrob("N", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                   just=c("left","top"),
                                                   gp=grid::gpar(col="black", fontsize=20)))

p15 <- gridExtra::arrangeGrob(p_dbscan_knn,
                              top = grid::textGrob("O", x = unit(0, "npc"), y   = unit(1, "npc"),
                                                   just=c("left","top"),
                                                   gp=grid::gpar(col="black", fontsize=20)))

p_final <- gridExtra::grid.arrange(p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,
                                   layout_matrix = rbind(c(1,2,3,5),c(1,2,4,5),
                                                         c(6,7,8,10),c(6,7,9,10),
                                                         c(11,12,13,15),c(11,12,14,15)))

################
# save results #
################
outfile_name_xls <- folder_image_str %>%
  paste("validation_outliers_cluster.xlsx", sep = "/")

result_lst <- list(
  normality_test_org = org_model_quality_df,
  normality_test_knn = knn_imp_model_quality_df,
  normality_test_median = median_imp_model_quality_df,
  normality_test_mean = mean_imp_model_quality_df,
  outlier_statistics = outliers_obj$outliersStatistics,
  rec_mse = mse_reconstruction_df,
  pca_variance = pca_variance_df,
  clustering_org = org_cluster_df,
  clustering_knn = knn_cluster_df,
  clustering_median = median_cluster_df,
  clustering_mean = mean_cluster_df,
  outliers = outliers_obj$outliers
)

result_lst %>%
  writexl::write_xlsx(outfile_name_xls)

data_lst <- list(
  original <- normalized_data$rawData,
  knn <- knn_imputed_data$rawData,
  median <- median_imputed_data$rawData,
  mean <- mean_imputed_data$rawData
)


outfile_name_data_xls <- folder_image_str %>%
  paste("validation_outliers_data.xlsx", sep = "/")
data_lst %>%
  writexl::write_xlsx(outfile_name_data_xls)

outfile_name_img <- folder_image_str %>%
  paste("validation_outliers_cluster.eps", sep = "/")

p_final %>%
  ggplot2::ggsave(filename = outfile_name_img,
                  units = "in",
                  width = 20,
                  height = 15)
